Package: CopernicusClimate
Title: Search Download and Handle Data from Copernicus Climate Data
        Service
Version: 0.0.6
Authors@R: 
    c(
      person("Pepijn", "de Vries", role = c("aut", "cre"),
             email = "pepijn.devries@outlook.com",
             comment = c(ORCID = "0000-0002-7961-6646"))
    )
Description: Subset and download data from EU Copernicus Climate Data Service:
    <https://cds.climate.copernicus.eu/>. Import information about the Earth's
    past, present and future climate from Copernicus into R without the need of
    external software.
Depends: R (>= 4.1.0)
Imports: dplyr, httr2, jsonlite, methods, purrr, rlang, stringr,
        tibble, tidyr
Suggests: clipr, ggplot2, knitr, RefManageR, rmarkdown, sf, stars,
        testthat (>= 3.0.0)
License: GPL (>= 3)
Encoding: UTF-8
Collate: 'CopernicusClimate-package.R' 'helpers.R' 'catalogue.R'
        'cite.R' 'jobs.R' 'licenses.R' 'python.R' 'retrieve.R'
        'stars.R' 'token.R'
Config/testthat/edition: 3
Language: en-GB
URL: https://pepijn-devries.github.io/CopernicusClimate/,
        https://github.com/pepijn-devries/CopernicusClimate/
BugReports: https://github.com/pepijn-devries/CopernicusClimate/issues
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-20 00:18:49 UTC; vries171
Author: Pepijn de Vries [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7961-6646>)
Maintainer: Pepijn de Vries <pepijn.devries@outlook.com>
Repository: CRAN
Date/Publication: 2026-07-20 12:20:51 UTC

# Generated by roxygen2: do not edit by hand

export(cds_accept_licence)
export(cds_accepted_licences)
export(cds_account_metrics)
export(cds_assign_star)
export(cds_build_request)
export(cds_catalogue_vocabulary)
export(cds_check_authentication)
export(cds_cite_dataset)
export(cds_dataset_form)
export(cds_delete_job)
export(cds_download_jobs)
export(cds_estimate_costs)
export(cds_get_account)
export(cds_get_token)
export(cds_job_results)
export(cds_list_datasets)
export(cds_list_jobs)
export(cds_list_licences)
export(cds_python_to_r)
export(cds_remove_star)
export(cds_search_datasets)
export(cds_set_token)
export(cds_starred)
export(cds_submit_job)
export(cds_token_works)
importFrom(rlang,.data)
importFrom(rlang,.env)

# CopernicusClimate <img src="man/figures/logo.svg" align="right" height="139" copyright="cc-sa" alt="logo" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/pepijn-devries/CopernicusClimate/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/pepijn-devries/CopernicusClimate/actions/workflows/R-CMD-check.yaml)
![cranlogs](https://cranlogs.r-pkg.org/badges/CopernicusClimate)
[![version](https://www.r-pkg.org/badges/version/CopernicusClimate)](https://CRAN.R-project.org/package=CopernicusClimate)
[![cran
checks](https://badges.cranchecks.info/worst/CopernicusClimate.svg)](https://cran.r-project.org/web/checks/check_results_CopernicusClimate.html)
[![CopernicusClimate status
badge](https://pepijn-devries.r-universe.dev/badges/CopernicusClimate)](https://pepijn-devries.r-universe.dev/CopernicusClimate)
[![Codecov test
coverage](https://codecov.io/gh/pepijn-devries/CopernicusClimate/graph/badge.svg)](https://app.codecov.io/gh/pepijn-devries/CopernicusClimate)
<!-- badges: end -->

## Overview

[The Copernicus Climate Change Service
(C3S)](https://climate.copernicus.eu/) has the mission of providing
information about the past, present and future climate, as well as tools
to enable climate change mitigation and adaptation strategies.

The C3S Climate Data Store provides open and state-of-the-art climate
data to scientists. This package allows users to download data from the
data store and handle it in R.

## Installation

Install CRAN release:

``` r
install.packages("CopernicusClimate")
```

Install latest developmental version from R-Universe:

``` r
install.packages("CopernicusClimate", repos = c('https://pepijn-devries.r-universe.dev', 'https://cloud.r-project.org'))
```

## Example

In order to download data from C3S you first need to submit a request
with `cds_submit_job()`. After your request has been processed by C3S,
you can download the data with `cds_download_jobs()`. This workflow is
demonstrated in the code snippet below. For a more detailed description
of the workflow see `vignette("download")`.

``` r
library(CopernicusClimate)
library(stars)   ## For loading spatial raster data
library(ggplot2) ## For plotting the data

if (cds_token_works()) { ## Make sure there is an operational access token
  
  ## Submit a download job:
  job <-
    cds_submit_job(
      "sis-agrometeorological-indicators",
      statistic = "day_time_mean",
      variable = "2m_temperature",
      year = "2025",
      month = "01",
      day = "01")
  
  ## Actually download the data:
  data_file <- cds_download_jobs(job$jobID, tempdir())
  
  ## Unzip the downloaded data:
  data_unzipped <- unzip(data_file$local, list = TRUE)
  unzip(data_file$local, exdir = tempdir())
  data_stars <- read_mdim(file.path(tempdir(), data_unzipped))
  
  ## Plot the downloaded data
  ggplot() +
    geom_stars(data = data_stars) +
    coord_sf() +
    labs(fill = "T(air 2m) [K]", x = NULL, y = NULL) +
    scale_fill_viridis_c(option = "inferno", na.value = "transparent")
}
```

<img src="man/figures/README-example-1.png" alt="Temperature map created from downloaded data" width="100%" />

## More of Copernicus

More R packages for exploring other Copernicus data services:

- [CopernicusMarine](https://github.com/pepijn-devries/CopernicusMarine)
  Dedicated to marine datasets

## Code of Conduct

Please note that the CopernicusClimate project is released with a
[Contributor Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tryCatch({
  Sys.setlocale("LC_ALL", "English")
})
library(ggplot2)
theme_set(theme_light())

## ----token--------------------------------------------------------------------
library(CopernicusClimate)

message(
  "The machine that rendered this vignette ",
  ifelse(
    cds_token_works(), "has", "does not have"),
  " a working token")


## ----get-licence, message=FALSE-----------------------------------------------
library(dplyr)

licence_info <-
  cds_dataset_form("reanalysis-era5-pressure-levels") |>
    filter(name == "licences")

licence_info <- licence_info$details[[1]]$details$licences[[1]]
print(licence_info)

## ----listing------------------------------------------------------------------
cds_list_datasets()

## ----search-------------------------------------------------------------------
cds_search_datasets(search = "rain", keywords = "Temporal coverage: Future")

## ----dataset-form-------------------------------------------------------------
dataset_form <-
  cds_dataset_form("reanalysis-era5-pressure-levels")

dataset_form

## ----possible-values----------------------------------------------------------
values <-
  dataset_form |>
  filter(name == "pressure_level") |>
  pull("details")

values[[1]]$details$values |> unlist()

## ----full-request-------------------------------------------------------------
request <- cds_build_request("reanalysis-era5-pressure-levels")
summary(request)

## ----specific-request---------------------------------------------------------
request <- cds_build_request(
  "reanalysis-era5-pressure-levels",
  variable       = "temperature",
  pressure_level = "1000",
  year           = "2025",
  month          = "01",
  day            = "01",
  area           = c(n = 60, w = -5, s = 40, e = 10),
  data_format    = "netcdf")
summary(request)

## ----estimate-full------------------------------------------------------------
if (cds_token_works()) {
  cds_estimate_costs("reanalysis-era5-pressure-levels")
} else {
  message("You need a working token to estimate costs")
}

## ----estimate-detailed--------------------------------------------------------
if (cds_token_works()) {
  cds_estimate_costs(request)
} else {
  message("You need a working token to estimate costs")
}

## ----submit, message=FALSE----------------------------------------------------
if (cds_token_works()) {
  job <-
    cds_submit_job(request)
  job
} else {
  message("You need a working token to submit a request")
}

## ----job-status---------------------------------------------------------------
if (cds_token_works()) {
  cds_list_jobs(job$jobID)
} else {
  message("You need a working token to get a job status")
}

## ----download, message=FALSE--------------------------------------------------
filename <- "result.nc"
if (cds_token_works()) {
  file_result <- cds_download_jobs(job$jobID, tempdir(), filename)
} else {
  message("Downloading data only works with a valid token")
}

## ----plot, fig.width=7, fig.height=3, message = FALSE, fig.alt="Plot created from downloaded data"----
fn <- file.path(tempdir(), filename)

if (file.exists(fn)) {
  
  library(stars)
  library(ggplot2)
  
  result <- read_mdim(fn)
  
  ggplot() +
    geom_stars(data = result) +
    coord_sf() +
    facet_wrap(~strftime(valid_time, "%H:%M"), nrow = 3) +
    scale_fill_viridis_c(option = "turbo") +
    labs(x = NULL, y = NULL, fill = "Temperature [K]") +
    theme(axis.text = element_blank())

} else {
  message("File wasn't downloaded")
}

#' @rdname cds_search_datasets
#' @include helpers.R
#' @family exploration-functions
#' @export
cds_list_datasets <- function(dataset, ...) {
  if (missing(dataset)) dataset <- NULL
  result <-
    .base_url |>
    paste("catalogue/v1/collections", dataset, sep = "/") |>
    .execute_request()
  
  if (is.null(result$collections)) .simplify(list(result)) else
    result$collections |>
      .simplify()
}

#' List or search Climate Data Service datasets
#' 
#' This will help you decide which datasets you wish to obtain.
#' @param dataset A specific dataset for which to list details. If missing
#' all datasets are listed.
#' @param search A string containing free text search terms to look for in the available datasets.
#' @param keywords A (vector of) string containing specific keywords. Should be listed in
#' `cds_catalogue_vocabulary()`
#' @param page When there are more search results than `limit`, results are paginated.
#' Use `page` to specify which page to return starting at `0` (default).
#' @param limit Use to limit the number of search results. Defaults to `50`.
#' @param ... Ignored
#' @returns A `data.frame` listing the datasets from the Climate Data Service. In
#' case of `cds_search_datasets()`, attributes named `SearchMeta` are added containing
#' the number of matching and number of returned datasets.
#' @examples
#' if (interactive()) {
#'   cds_list_datasets()
#'   cds_search_datasets("rain", "Temporal coverage: Future")
#' }
#' @include helpers.R
#' @family exploration-functions
#' @export
cds_search_datasets <- function(search, keywords, page = 0, limit = 50, ...) {
  if (missing(search)) search <- ""
  if (missing(keywords)) keywords <- list()
  keywords <- unlist(keywords)
  result <-
    .base_url |>
    paste0("/catalogue/v1/datasets", sep = "") |>
    .execute_request("", "POST",
                     list(q = search, kw = as.list(keywords), idx = list(), sortby = "update",
                          page = page, limit = limit, search_stats = TRUE))
  
  meta <- list(numberMatched = result$numberMatched,
               numberReturned = result$numberReturned)
  
  result <-
    result$collections |>
    .simplify()
  attributes(result)$SearchMeta <- meta
  result
}

#' Obtain an overview of options to subset a dataset
#' 
#' This function provides an overview of parameters that can be used to subset a
#' dataset. It can help you set up a request with `cds_submit_job()` or `cds_build_request()`.
#' @param dataset A name of a dataset to explore
#' @param ... Ignored
#' @returns A `data.frame` with aspects of the dataset that can be subsetted, when defining
#' a job. See also `cds_submit_job()` and `cds_build_request()`
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_dataset_form("reanalysis-era5-pressure-levels")
#' }
#' @family helper-functions
#' @export
cds_dataset_form <- function(dataset, ...) {
  result <-
    .base_url |>
    paste("catalogue/v1/collections", dataset, "form.json", sep = "/") |>
    .execute_request() |>
    purrr::map_df(~ .x)
  if ("details" %in% names(result))
    result <- tidyr::nest(result, details = "details")
  if ("children" %in% names(result))
    result <- tidyr::nest(result, children = "children")
  result
}

#' List catalogue vocabulary
#' 
#' The catalogue uses a specific vocabulary for keywords. This function
#' produces an overview.
#' @param ... Ignored
#' @returns Returns a `data.frame` of keyword vocabulary used by the catalogue.
#' @examples
#' if (interactive()) {
#'   cds_catalogue_vocabulary()
#' }
#' @include helpers.R
#' @family exploration-functions
#' @export
cds_catalogue_vocabulary <- function(...) {
  result <-
    .base_url |>
    paste0("/catalogue/v1/vocabularies/keywords", sep = "") |>
    .execute_request()
  result$keywords |> .simplify()
}

test_that("Listing datasets results in at least 100 hits", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    nrow(cds_list_datasets()) >= 100
  })
})

test_that("Searching rain forecasts results in exactly that", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    result <- cds_search_datasets("rain", "Temporal coverage: Future")
    lapply(result$keywords, \(x) "Temporal coverage: Future" %in% x) |>
      unlist() |>
      all()
  })
})

test_that("Limiting a search works", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    nrow(cds_search_datasets(limit = 5)) == 5
  })
})

test_that("Getting catalogue vocabulary works", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    "Temporal coverage: Future" %in% cds_catalogue_vocabulary()$id
  })
})

#' List accepted licences
#' 
#' In order to use specific features of the Climate Data Service, you may first
#' need to accept specific licences. This function will list all licenses you
#' have accepted.
#' @param scope Scope of the licenses to be listed should be one of `"all"` (default),
#' `"dataset"` or `"portal"`.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns Returns a `data.frame` listing the accepted licenses.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_accepted_licences("portal")
#' }
#' @include helpers.R
#' @family licences
#' @export
cds_accepted_licences <- function(
    scope = c("all", "dataset", "portal"),
    ...,
    token = cds_get_token()) {
  scope <- match.arg(scope)
  .base_url |>
    paste0("/profiles/v1/account/licences", sep = "") |>
    httr2::url_modify_query(scope = scope) |>
    .execute_request(token) |>
    dplyr::first() |>
    lapply(data.frame) |>
    dplyr::bind_rows()
}

#' Accept a specific licence
#' 
#' Call this function if you wish to accept a specific license.
#' @param license The license id you wish to accept.
#' @param revision The revision number of the license you are accepting.
#' Should always be the latest revision.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns Returns a `data.frame` containing the accepted license
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_accept_licence("cc-by", 1)
#' }
#' @include helpers.R
#' @family licences
#' @export
cds_accept_licence <- function(
    license, revision, ..., token = cds_get_token()) {
  .base_url |>
    paste("profiles/v1/account/licences", license, sep = "/") |>
    .make_request(token, "PUT") |>
    httr2::req_body_json(list(revision = revision)) |>
    httr2::req_error(body = .req_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json() |>
    dplyr::as_tibble()
}

#' Obtain a list of licences that can be accepted
#' 
#' Some datasets require you to accept specific licences. This function will provide
#' an overview of all licences associated with datasets and can be accepted.
#' @param ... Ignored
#' @returns Returns a `data.frame` of available licences that can be accepted.
#' @examples
#' if (interactive()) {
#'   cds_list_licences()
#' }
#' @family licences
#' @export
cds_list_licences <- function(...) {
  result <-
    .base_url |>
    paste("catalogue/v1/vocabularies/licences", sep = "/") |>
    .execute_request()
  .simplify(result$licences)
}

test_that("Accepted licences can be obtained", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_no_error({
    cds_accepted_licences()
  })
})

test_that("A license can be accepted", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_no_error({
    cds_accept_licence("cc-by", 1)
  })
})

test_that("Licenses can be listed", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    nrow(cds_list_licences()) > 10
  })
})

#' Submit a download job for a dataset
#' 
#' Submit a request to the Copernicus Climate Data Service to download
#' (part of) a dataset. If the request is successful, a job identifier is
#' returned which can be used to actually download the data (`cds_download_jobs()`).
#' 
#' @param dataset The dataset name to be downloaded, or a `list` returned by
#' `cds_build_request()`. When this argument is missing the function will
#' attempt to build a request with text on the clipboard using `cds_python_to_r()`.
#' @param ... Subsetting parameters passed onto `cds_build_request()`. Should
#' be empty when `dataset` is missing or a `list`.
#' @param wait A `logical` value indicating if the function should wait for the
#' submitted job to finish. Set it to `FALSE` to continue without waiting
#' @param check_quota Each account has a quota of data that can be downloaded.
#' If this argument is set to `TRUE` (default) it is checked if the request doesn't
#' exceed your quota. Set it to `FALSE` to skip this step and speed up the submission.
#' @param check_licence Datasets generally require you to accept certain terms of use.
#' If this argument is set to `TRUE` (default), it will be checked if you have accepted all
#' required licences for the submitted request. Set it to `FALSE` to skip this step and
#'  speed up the submission.
#' @inheritParams cds_check_authentication
#' @returns Returns a `data.frame` containing information about the submitted job.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   job <- cds_submit_job(
#'       dataset        = "reanalysis-era5-pressure-levels",
#'       variable       = "geopotential",
#'       product_type   = "reanalysis",
#'       area           = c(n = 55, w = -1, s = 50, e = 10),
#'       year           = "2024",
#'       month          = "03",
#'       day            = "01",
#'       pressure_level = "1000",
#'       data_format    = "netcdf"
#'     )
#'
#'   ## Or split into two separate steps:
#'   
#'   req <- cds_build_request(
#'       dataset        = "reanalysis-era5-pressure-levels",
#'       variable       = "geopotential",
#'       product_type   = "reanalysis",
#'       area           = c(n = 55, w = -1, s = 50, e = 10),
#'       year           = "2024",
#'       month          = "03",
#'       day            = "01",
#'       pressure_level = "1000",
#'       data_format    = "netcdf"
#'     )
#'   job <- cds_submit_job(req)
#' }
#' @include helpers.R
#' @family job-functions
#' @export
cds_submit_job <- function(
    dataset, ..., wait = TRUE, check_quota = TRUE, check_licence = TRUE,
    token = cds_get_token()) {
  #https://cds.climate.copernicus.eu/api/retrieve/v1/docs
  message("Building request")
  if (missing(dataset)) dataset <- NULL
  
  form <- cds_build_request(dataset, ...)
  if (is.null(dataset) || is.list(dataset)) dataset <- attr(form, "dataset")
  if (check_quota) {
    message("Checking quota")
    quota <- .cds_estimate_costs(form, token = token)
    if (quota$cost > quota$limit) {
      rlang::abort(c(x = sprintf("This request (%i) exceeds your quota (%i)",
                                 quota$cost, quota$limit),
                     i = "Try narrowing your request"))
    }
  }
  if (is.null(attributes(form)$licences)) {
    message("Required license unknown, skipping check")
  } else if (check_licence) {
    message("Checking license")
    missing_licence <- attributes(form)$licences$id
    missing_licence <- missing_licence[!missing_licence %in% cds_accepted_licences()$id]
    if (length(missing_licence) > 0) {
      rlang::abort(c(
        x = sprintf("Dataset requires you to accept these licences: %s",
                    paste(missing_licence, collapse = ", ")),
        i = "Use `cds_accept_licence()` to accept the required licences and try again"
      ))
    }
  }
  message("Submitting job")
  job <-
    .base_url |>
    paste("retrieve/v1/processes", dataset, "execution", sep = "/") |>
    .execute_request(token, "POST", list(inputs = form)) |>
    list() |>
    .simplify()
  
  if (wait) {
    wait_anim <- c("-", "\\", "|", "/")
    i <- 1
    repeat {
      job <- cds_list_jobs(job$jobID, token = token)
      if ("finished" %in% names(job)) {wait_anim <- " "; i <- 1}
      message(paste("\rChecking job status:", wait_anim[i], job$status, "    "),
              appendLF = FALSE)
      if ("finished" %in% names(job)) {
        message("")
        break
      }
      i <- i + 1
      if (i > length(wait_anim)) i <- 1
      Sys.sleep(1)
    }
  }
  return (job)
}

#' Prepare a request for downloading a dataset
#' 
#' This function is used by `cds_estimate_costs()` and `cds_submit_job()`
#' to subset a dataset before downloading. It will also help you to explore
#' which parameters are available for subsetting.
#' @param dataset The dataset name to be used for setting up a request.
#' @param ... Parameters for subsetting the dataset. Use `cds_dataset_form()` to inquiry
#' which parameters and parameter values are available for a specific dataset.
#' If left blank it will take default parameter values.
#' @returns Returns a named list, which can be used to submit a job (`cds_submit_job()`)
#' or inquiry its cost (`cds_estimate_costs()`).
#' @examples
#' if (interactive()) {
#'   cds_build_request(
#'     dataset        = "reanalysis-era5-pressure-levels",
#'     variable       = "geopotential",
#'     product_type   = "reanalysis",
#'     area           = c(n = 55, w = -1, s = 50, e = 10),
#'     year           = "2024",
#'     month          = "03",
#'     day            = "01",
#'     pressure_level = "1000",
#'     data_format    = "netcdf"
#'   )
#' }
#' @family helper-functions
#' @export
cds_build_request <- function(dataset, ...) {
  form_result <- list(...)
  if ((missing(dataset) || is.null(dataset)) && length(form_result) == 0) {
    form_result <- cds_python_to_r()
    dataset <- attr(form_result, "dataset")
  } else if (is.list(dataset) && length(form_result) == 0) {
    ## Assume that if dataset is a list and there are no
    ## further arguments, 'dataset' is already parsed by
    ## cds_build_request and can be returned as is.
    return (dataset)
  }

  constraints_all <- .cds_constraints(dataset,  structure(list(), names = character(0)))
  form <- cds_dataset_form(dataset) |>
    dplyr::mutate(
      required = ifelse(is.na(.data$required), FALSE, .data$required)
    )

  known <- names(form_result) %in% form$name
  unknown <- names(form_result)[!known]
  for (nm in names(form_result)) {
    if (nm %in% unknown) {
      message("Removing unknown field ", nm)
      form_result[[nm]] <- NULL
    } else {
      details <- form |> dplyr::filter(.data$name == nm)
      details <- details$details[[1]]$details
      if (!is.null(details$groups)) {
        my_groups <- lapply(details$groups, `[[`, "label") |> unlist() |>
          tolower() |> stringr::str_replace_all(" ", "_")
        sel <- my_groups %in% form_result[[nm]]
        if (any(sel)) {
          details$values <-
            lapply(details$groups[sel], `[[`, "values") |>
            unlist() |>
            unique() |>
            list()
          form_result[[nm]] <- details$values |> unlist()
        }
      }
      if (!is.null(details$values) &&
          any(!form_result[[nm]] %in% unlist(details$values))) {
        rlang::abort(c(x = sprintf("Unknown value '%s'", paste(unlist(form_result[[nm]]),
                                                               collapse = ", ")),
                       i = paste("Expected one of", paste(unlist(details$values),
                                                          collapse = ", "))))
      }
      if (!is.null(details$values)) {
        form_result[[nm]] <- methods::as(unlist(form_result[[nm]]),
                                         typeof(details$values[[1]]))
      }
    }
  }
  required_elements <- form$name[form$required]
  not_included <- required_elements[!required_elements %in% names(form_result)]
  
  # Check lengths
  for (nm in names(form_result)) {
    details <- form |> dplyr::filter(.data$name == nm)
    type <- details$type
    switch(
      type,
      StringListWidget = {
        form_result[[nm]] <- as.list(unlist(form_result[[nm]]))
      },
      StringListArrayWidget = {
        form_result[[nm]] <- as.list(unlist(form_result[[nm]]))
      },
      StringChoiceWidget = {
        if (length(form_result[[nm]]) != 1) {
          rlang::abort(c(x = sprintf("Found multiple values for field '%s'", nm),
                         i = "Select only one value and try again"))
        } else {
          form_result[[nm]] <- unlist(form_result[[nm]])
        }
      },
      GeographicExtentWidget = {
        geo_details <- details$details[[1]]$details
        current <- unlist(form_result[[nm]])
        if (inherits(current, "bbox")) {
          if (requireNamespace("sf")) {
            if (!is.na(sf::st_crs(form_result[[nm]]))) {
              current <- sf::st_transform(form_result[[nm]], 4326)
            }
            current <- c(n = current[["ymax"]],
                         w = current[["xmin"]],
                         s = current[["ymin"]],
                         e = current[["xmax"]])
            
          } else {
            rlang::abort(
              c(x = "Using a `bbox` as `area` arguments needs package `sf`",
                i = "Install package `sf` and try again")
            )
          }
        }
        if (length(current) != 4)
          rlang::abort(c(x = sprintf("Expected rectangular bounding box, with 4 values (%s)",
                                     paste(names(geo_details$default), collapse = ", ")),
                         i = "Provide a correct bounding box"))
        if (is.null(names(current))) names(current) <- c("n", "w", "s", "e")
        if (current["s"] > current["n"] || current["e"] < current["w"])
          rlang::abort(c(x = "North should be larger than South. East should be larger than West",
                         i = "Check your bounding box for correctness"))
        if (current[["n"]] > geo_details$range$n ||
            current[["s"]] < geo_details$range$s ||
            current[["e"]] > geo_details$range$e ||
            current[["w"]] < geo_details$range$w)
          rlang::abort(c(x = "Coordinates of `area` out of range",
                         i = sprintf("Check if your coordinates are in range (%s)",
                                     paste(unlist(geo_details$range), collapse = ", "))))
        form_result[[nm]] <- current
      },
      LicenceWidget = {
        form_result[[nm]] <- NULL
      },
      rlang::abort(c(x = paste("Unknown field type", type),
                     i = "Please report at <https://github.com/pepijn-devries/CopernicusClimate/issues> with a reprex"))
    )
  }

  if (length(form_result) == 0) names(form_result) <- character(0)
  constraints <- .cds_constraints(dataset, form_result)
  for (missing_element in not_included) {
    if (!is.null(constraints[[missing_element]])) {
      details <-
        form |>
        dplyr::filter(.data$name == missing_element)
      type <- details$type
      details <- details$details[[1]]$details
      if (!is.null(details$default)) {
        form_result[[missing_element]] <- details$default |> unlist() |> unique()
      } else {
        dat <- constraints[[missing_element]]
        if (type %in% "StringChoiceWidget") dat <- dat[[1]]

        if (length(dat) == 0) form_result[[missing_element]] <- NULL else
          form_result[[missing_element]] <- dat
      }
    }
  }
  
  licences <- form |>
    dplyr::filter(.data$name == "licences") |>
    dplyr::pull("details")
  licences <- licences[[1]]$details$licences |> .simplify()
  attributes(form_result)$licences <- licences
  attributes(form_result)$dataset  <- dataset
  return( form_result )
}

.cds_constraints <- function(dataset, form) {
  .base_url |>
    paste("retrieve/v1/processes", dataset, "constraints", sep = "/") |>
    .execute_request(token = "", "POST", list(inputs = form))
}

#' Check the cost of a request against your quota
#' 
#' Each account has a limit to the amount of data that can be downloaded.
#' Use this function to check if a request exceeds your quota.
#' @param dataset A dataset name to be inspected
#' @param ... Parameters passed on to `cds_build_request()`
#' @inheritParams cds_check_authentication
#' @returns Returns a named list indicating the available quota and
#' the estimated cost for a request specified with `...`-arguments.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_estimate_costs(
#'     dataset        = "reanalysis-era5-pressure-levels",
#'     variable       = "geopotential",
#'     product_type   = "reanalysis",
#'     area           = c(n = 55, w = -1, s = 50, e = 10),
#'     year           = "2024",
#'     month          = "03",
#'     day            = "01",
#'     pressure_level = "1000",
#'     data_format    = "netcdf"
#'   )
#'   
#'   cds_estimate_costs(dataset = "reanalysis-era5-pressure-levels")
#' }
#' @include helpers.R
#' @family helper-functions
#' @export
cds_estimate_costs <- function(dataset, ..., token = cds_get_token()) {
  if (missing(dataset)) dataset <- NULL
  .cds_estimate_costs(dataset, ..., token = token)
}

.cds_estimate_costs <- function(dataset, ..., token) {
  form <- cds_build_request(dataset, ...)
  if (is.null(dataset) || is.list(dataset)) dataset <- attr(form, "dataset")
  .base_url |>
    paste("retrieve/v1/processes", dataset, "costing", sep = "/") |>
    .execute_request(token, "POST", list(inputs = form))
}

#' Download specific jobs
#' 
#' After submitting one or more jobs with `cds_submit_job()`, you can download the resulting
#' files with `cds_download_jobs()`. See `vignette("download")` for more details.
#' @param job_id If a specific job identifier is listed here, only the files resulting
#' from those jobs are downloaded. If left blank, all successful jobs are downloaded.
#' @param destination Destination path to store downloaded files.
#' @param names File names for the downloaded files. If missing, the cryptic hexadecimal
#' file name is taken from the job.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns A `data.frame` of all downloaded files. Contains a column `local` with the path
#' to the locally stored files.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   job <- cds_submit_job(
#'       dataset        = "reanalysis-era5-pressure-levels",
#'       variable       = "geopotential",
#'       product_type   = "reanalysis",
#'       area           = c(n = 55, w = -1, s = 50, e = 10),
#'       year           = "2024",
#'       month          = "03",
#'       day            = "01",
#'       pressure_level = "1000",
#'       data_format    = "netcdf"
#'     )
#'   cds_download_jobs(job$jobID, tempdir())
#' }
#' @include helpers.R
#' @family job-functions
#' @export
cds_download_jobs <- function(job_id, destination, names, ..., token = cds_get_token()) {
  if (!missing(names) && length(job_id) != length(names))
    stop("Argument `names` should have the same length as `job_id`.")
  missing_names <- missing(names)
  repeat {
    
    jobs <- cds_list_jobs(job_id, limit = 1000, token = token)
    busy_jobs <- jobs$status %in% c("accepted", "running")
    if (any(busy_jobs)) {
      message("\rWaiting for ", sum(busy_jobs), " job(s) to complete   ", appendLF = FALSE)
    } else break
    Sys.sleep(1)
  }
  message("")
  jobs <-
    cds_job_results(job_id, token = token) |>
    dplyr::mutate(
      name = basename(.data$href),
      success = !is.na(.data$href))
  if (any(!jobs$success)) message("Skipping ", sum(!jobs$success), " unsuccessful jobs")
  jobs <- jobs |> dplyr::filter(.data$success)
  if (nrow(jobs) == 0) stop("No successful jobs found to download")
  dupes <- duplicated(jobs$href)
  if (any(dupes)) message("Skipped ", sum(dupes), " identical files")
  jobs <- jobs |> dplyr::filter(!.env$dupes)
  requests <-
    lapply(jobs$href, httr2::request)
  message("Start downloading files:")
  fnames <- if (missing_names) jobs$name else names
  responses <- httr2::req_perform_parallel(
    requests,
    paths = file.path(destination, fnames)
  )
  jobs |>
    dplyr::mutate(
      responses = responses,
      local = lapply(responses, \(x) x$body |> unclass()) |> unlist()
    )
}

test_that("Download workflow works", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_true({
    job <- cds_submit_job(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = c(n = 55, w = -1, s = 50, e = 10),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf"
    ) |>
      suppressMessages()
    my_download <-
      cds_download_jobs(job$jobID, tempdir()) |>
      suppressMessages()
    file.size(my_download$local) > 0
  })
})

test_that("Area can be a `bbox` class object", {
  skip_if_offline()
  skip_on_cran()
  expect_no_error({
    cds_build_request(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = sf::st_bbox(c(ymax = 55, xmin = -1,
                                     ymin = 50, xmax = 10)),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf"
    )
  })
})

test_that("Costs can be estimated", {
  skip_on_cran()
  skip_if_offline()
  expect_no_error({
    cds_estimate_costs("reanalysis-era5-pressure-levels")
  })
})

test_that("Job can be cancelled", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_true({
    job <- cds_submit_job(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = c(n = 55, w = -1, s = 50, e = 10),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf",
      wait           = FALSE
    ) |>
      suppressMessages()
    job_del <- cds_delete_job(job$jobID)
    job_del$status == "dismissed"
  })
})

test_that("Job can listed", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_no_error({
    cds_list_jobs(status = "accepted")
  })
})

test_that("Duplicated default values are removed", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    request <-
      cds_build_request(
        "reanalysis-era5-single-levels",
        variable = "v_component_stokes_drift",
        year = "2025",
        month = "01",
        day = "01",
        data_format="netcdf"
      )
    length(request$product_type) == 1
  })
})

test_that("Grouped variables can also be used to build query", {
  skip_on_cran()
  skip_if_offline()
  expect_true({
    req <-
      cds_build_request(
        "reanalysis-era5-single-levels",
        variable = "wind" ## <- this is a group
      )
    ## should return all variables within group
    length(req$variable) > 1
  })
})

python_code <-
"import cdsapi

dataset = \"reanalysis-era5-pressure-levels\"
request = {
  \"product_type\": [\"reanalysis\"],
  \"variable\": [\"geopotential\"],
  \"year\": [\"2024\"],
  \"month\": [\"03\"],
  \"day\": [\"01\"],
  \"time\": [
    \"00:00\", \"01:00\", \"02:00\",
    \"03:00\", \"04:00\", \"05:00\",
    \"06:00\", \"07:00\", \"08:00\",
    \"09:00\", \"10:00\", \"11:00\",
    \"12:00\", \"13:00\", \"14:00\",
    \"15:00\", \"16:00\", \"17:00\",
    \"18:00\", \"19:00\", \"20:00\",
    \"21:00\", \"22:00\", \"23:00\"
  ],
  \"pressure_level\": [\"1000\"],
  \"data_format\": \"netcdf\",
  \"download_format\": \"unarchived\",
  \"area\": [55, -1, 50, 10]
}

client = cdsapi.Client()
client.retrieve(dataset, request).download()"

test_that("Clipboard can be used to define request", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(clipr::clipr_available(), "This test needs access to the clipboard")
  skip_if_not(cds_token_works(), "This test requires a working access token")
  expect_no_error({
    clipr::write_clip(python_code)
    cds_submit_job() |>
      suppressMessages()
  })
})

test_that("Python request can be converted to R request", {
  skip_on_cran()
  skip_if_offline()
  expect_no_error({
    req <- cds_python_to_r(python_code)
  })
})

Sys.setenv(CLIPR_ALLOW = TRUE)

test_that("Request cannot exceed quota", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_submit_job("reanalysis-era5-pressure-levels") |>
      suppressMessages()
  }, regexp = "*.?This request \\(\\d+\\) exceeds your quota*.?")
})

test_that("Only valid values can be submitted", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_submit_job("reanalysis-era5-pressure-levels", pressure_level = -1) |>
      suppressMessages()
  }, regexp = "*.?Unknown value*.?")
})

test_that("Unknown fields are removed", {
  skip_on_cran()
  skip_if_offline()
  expect_message({
    cds_build_request("reanalysis-era5-pressure-levels", foo = "bar")
  }, regexp = "Removing unknown field foo")
})

test_that("Cannot provide multiple values when expecting one", {
  skip_on_cran()
  skip_if_offline()
  expect_error({
    cds_build_request("reanalysis-era5-pressure-levels",
                      data_format = c("grib", "netcdf")) |>
      suppressMessages()
  }, regexp = "*.?Found multiple values for field*.?")
})

test_that("Names cannot have a different length than jobs", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_download_jobs(1:3, names = "foo")
  }, "should have the same length")
})

test_that("Cannot delete non-existing stars", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_remove_star("foobar")
  }, regexp = "*.?Cannot find dataset foobar*.?")
})

test_that("Cannot pass invalid parameters to url", {
  skip_on_cran()
  skip_if_offline()
  expect_error({
    cds_search_datasets(page = "foobar")
  }, regexp = "*.?Input should be a valid integer*.?")
})

test_that("Bounding box with no 4 values is rejected", {
  skip_on_cran()
  skip_if_offline()
  expect_error({
    cds_build_request("reanalysis-era5-single-levels", area = 1:3)
  }, "4 values")
})

test_that("Invalid bounding box coords are rejected", {
  skip_on_cran()
  skip_if_offline()
  expect_error({
    cds_build_request("reanalysis-era5-single-levels", area = 1:4)
  }, "correctness")
})

test_that("Coordinates out of range are rejected", {
  skip_on_cran()
  skip_if_offline()
  expect_error({
    cds_build_request("reanalysis-era5-single-levels", area = c(95, -1, 94, 0))
  }, "out of range")
})

test_that("Cost estimation fails with nothing on clipboard", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_estimate_costs() |> suppressMessages()
  }, "System clipboard contained no readable tex|Clipboard on X11|missing required 'clipr'|Failed to convert text to CDSAPI request")
})

test_that("Submission fails with nothing on clipboard", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_submit_job() |> suppressMessages()
  }, "System clipboard contained no readable text|Clipboard on X11|missing required 'clipr'|Failed to convert text to CDSAPI request")
})

test_that("Building request fails with nothing on clipboard", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_error({
    cds_build_request() |> suppressMessages()
  }, "System clipboard contained no readable text|Clipboard on X11|missing required 'clipr'|Failed to convert text to CDSAPI request")
})

#' List jobs submitted to the Climate Data Service
#' 
#' Once submitted with `cds_submit_job()` you can check the status of the job with
#' this function. You can list all available jobs, or specific jobs.
#' @param job_id The id of a specific job, if you want the results for that job. If `NULL`
#' (default) it is ignored.
#' @param status Only return jobs with the status stated by this argument. Default is `NULL`
#' meaning that jobs with any status are returned. Should be any of `"accepted"`, `"running"`,
#' `"successful"`, `"failed"`, or `"rejected"`.
#' @param limit Use to limit the number of listed results. Defaults to `50`.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns Returns a `data.frame` of submitted jobs.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_list_jobs()
#' }
#' @include helpers.R
#' @family job-functions
#' @export
cds_list_jobs <- function(job_id = NULL, status = NULL, limit = 50,
                          ..., token = cds_get_token()) {
  # https://cds.climate.copernicus.eu/api/retrieve/v1/docs
  if (!is.null(status))
    status <- match.arg(
      status, c("accepted", "running", "successful", "failed", "rejected"))
  
  job_info <- function(id, stat, lim) {
    result <-
      .base_url |>
      paste("retrieve/v1/jobs", id, sep = "/") |>
      httr2::url_modify_query(status = stat, limit = lim) |>
      .execute_request(token)
    
    meta <- result$metadata
    result <-
      if ("jobs" %in% names(result)) {
        result$jobs |>
          .simplify()
      } else {
        result |> list() |> .simplify()
      }
    attributes(result)$MetaList <- meta
    result
  }
  
  if (is.null(job_id)) job_info(NULL, status, limit) else
    lapply(job_id, job_info, status, limit) |>
      dplyr::bind_rows()
}

#' Delete/cancel jobs submitted to the Climate Data Service
#' 
#' When you regret submitting a job, you can cancel it by calling this function.
#' @param job_id Hexadecimal code used as identifier of a job. Identifies the job
#' to be cancelled.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns Returns a `data.frame` with information about the cancelled job.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   job <- cds_submit_job(
#'       dataset        = "reanalysis-era5-pressure-levels",
#'       variable       = "geopotential",
#'       product_type   = "reanalysis",
#'       area           = c(n = 55, w = -1, s = 50, e = 10),
#'       year           = "2024",
#'       month          = "03",
#'       day            = "01",
#'       pressure_level = "1000",
#'       data_format    = "netcdf"
#'     )
#'   cds_delete_job(job$jobID, tempdir())
#' }
#' @include helpers.R
#' @family job-functions
#' @export
cds_delete_job <- function(job_id, ..., token = cds_get_token()) {
  result <-
    .base_url |>
    paste("retrieve/v1/jobs/delete", sep = "/") |>
    .execute_request(token, "POST", list(job_ids = as.list(job_id)))
  result$jobs |> .simplify()
}

#' Get the results for a submitted job
#' 
#' When a job is completed, you can get its results (i.e. download link)
#' by calling this function.
#' @param job_id Hexadecimal code used as identifier of a job. Identifies the job
#' for which to obtain the results.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns Returns a `data.frame` with information about the requested job.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   job <- cds_submit_job(
#'       dataset        = "reanalysis-era5-pressure-levels",
#'       variable       = "geopotential",
#'       product_type   = "reanalysis",
#'       area           = c(n = 55, w = -1, s = 50, e = 10),
#'       year           = "2024",
#'       month          = "03",
#'       day            = "01",
#'       pressure_level = "1000",
#'       data_format    = "netcdf"
#'     )
#'   cds_job_results(job$jobID)
#' }
#' @include helpers.R
#' @family job-functions
#' @export
cds_job_results <- function(job_id, ..., token = cds_get_token()) {
  result <-
    .base_url |>
    paste("retrieve/v1/jobs/%s/results", sep = "/") |>
    sprintf(job_id) |>
    lapply(
      .execute_request, token, "GET", list(job_ids = as.list(job_id))) |>
    lapply(`[[`, "asset") |>
    lapply(`[[`, "value") |>
    lapply(dplyr::as_tibble) |>
    lapply(\(x) dplyr::bind_cols(dplyr::tibble(job_id = NA), x)) |>
    dplyr::bind_rows() |>
    dplyr::mutate(job_id = .env$job_id)
}

#' Add or remove a star to a dataset, or list starred datasets
#' 
#' Use stars to keep track of your favourite datasets. Use these
#' functions to add or remove a star.
#' @param dataset Name of the dataset to assign a star to, or remove it from.
#' @param ... Ignored
#' @inheritParams cds_check_authentication
#' @returns In case of `cds_assign_star()` returns the name of the starred
#' dataset. In case of `cds_starred()` a vector of names of starred datasets.
#' In case of `cds_remove_star()` returns `NULL` invisibly.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_assign_star("reanalysis-carra-single-levels")
#'   cds_starred()
#'   cds_remove_star("reanalysis-carra-single-levels")
#' }
#' @include helpers.R
#' @family exploration-functions
#' @export
cds_starred <- function(..., token = cds_get_token()) {
  account <- cds_get_account(token = token)
  account$starred_datasets |> unlist()
}

#' @rdname cds_starred
#' @include helpers.R
#' @family exploration-functions
#' @export
cds_assign_star <- function(dataset, ..., token = cds_get_token()) {
  .base_url |>
    paste0("/profiles/v1/account/starred", sep = "") |>
    .execute_request(token, "POST", list(uid = dataset)) |>
    unlist()
}

#' @rdname cds_starred
#' @include helpers.R
#' @family exploration-functions
#' @export
cds_remove_star <- function(dataset, ..., token = cds_get_token()) {
  .base_url |>
    paste("profiles/v1/account/starred", dataset, sep = "/") |>
    .make_request(token, "DELETE") |>
    httr2::req_error(body = .req_error) |>
    httr2::req_perform()
  ## removing a star returns empty body
  invisible(NULL)
}

test_that("Datasets can be starred", {
  skip_on_cran()
  skip_if(Sys.getenv("SKIPSTARS") == "TRUE",
          "Skip to avoid too many simultaneous star operations")
  skip_if_not(cds_token_works())
  skip_if_offline()
  expect_false({
    nm <- "reanalysis-era5-pressure-levels"
    cds_assign_star(nm)
    Sys.sleep(2) ## Wait to make sure request has been processed
    cds_remove_star(nm)
    Sys.sleep(2) ## Wait to make sure request has been processed
    nm %in% cds_starred()
  })
})

#' Get or set a Climate Data Service API key
#' 
#' Many of the Climate Data Services features require a personal
#' Application Programming Interface (API) key. This function
#' will get a token that was previously stored (`cds_set_token()`),
#' and can be used throughout the R session.
#' 
#' To use an API key, you first need to get one from the
#' Climate Data Service. You can do so, by creating an account
#' and then initialise the key at <https://cds.climate.copernicus.eu/profile>.
#' 
#' There are different locations where the key can be stored and where
#' `cds_get_token()` will look. It will look for the first successful value of (in order):
#' the environment variable named `"CDSAPI_KEY"`, the R `getOption()` named `"CDSAPI_KEY"`,
#' the environment variable named `"ECMWF_DATASTORES_KEY"`, and the R `getOption()` named
#' `"ECMWF_DATASTORES_KEY"`.
#'
#' You can set the key at the start of each R session with `cds_set_token()`. If you
#' want a persistent solution, you can add the environment variable (with names shown above)
#' to your system. Or you can add the option (with the names shown above) to your
#' `".profile"` file. This will help you obscure your sensitive account information in your R script.
#'
#' @param token The API key you wish to set as an R option or to an environment variable.
#' @param method Method to store the API key. Should be either `"option"` (default) or
#' `"sysenv"`.
#' @param ... Ignored
#' @returns `cds_get_token()` will return an API key token if it has been set. If
#' it is not set it will return an empty string: `""`. `cds_set_token()` will
#' return `NULL` invisibly.
#' @examples
#' if (interactive() && !cds_token_works()) {
#'   cds_set_token("this-is-a-dummy-token", "option")
#' }
#' 
#' if (interactive()) {
#'   cds_get_token()
#' }
#' @include helpers.R
#' @family authentication-functions
#' @export
cds_get_token <- function(...) {
  token <- Sys.getenv("CDSAPI_KEY")
  if (token == "")    token <- getOption("CDSAPI_KEY")
  if (is.null(token)) token <- Sys.getenv("ECMWF_DATASTORES_KEY")
  if (token == "")    token <- getOption("ECMWF_DATASTORES_KEY")
  if (is.null(token)) token <- ""
  return(token)
}

#' @rdname cds_get_token
#' @export
cds_set_token <- function(token, method = c("option", "sysenv"), ...) {
  method <- match.arg(method)
  switch(
    method,
    option = {
      options(CDSAPI_KEY = token)
    },
    sysenv = {
      Sys.setenv(CDSAPI_KEY = token)
    })
  return(invisible())
}

#' Check if authentication works with a specific token
#' 
#' Checks if the specified API key can be used for authentication at
#' the Climate Data Service.
#' @param token An API key to be used for authentication. Will use
#' `cds_get_token()` by default.
#' @param ... Ignored
#' @returns `cds_check_authentication()` will return
#' some account information when successful but throws an error if
#' it is not. In contrast `cds_token_works()` returns a `logical`
#' value and will not throw an error upon failure.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_check_authentication()
#' }
#' @include helpers.R
#' @family authentication-functions
#' @export
cds_check_authentication <- function(token = cds_get_token(), ...) {
  #https://cds.climate.copernicus.eu/api/profiles/v1/docs
  .base_url |>
    paste0("/profiles/v1/account/verification/pat", sep = "") |>
    .execute_request(token, "POST")
}

#' @rdname cds_check_authentication
#' @include helpers.R
#' @export
cds_token_works <- function(token = cds_get_token(), ...) {
  result <-
    tryCatch({
      cds_check_authentication(token)
      TRUE
    }, error = function(e) FALSE)
  return(result)
}

#' Get account details
#' 
#' Retrieve account details associated with the provided `token`
#' @inheritParams cds_check_authentication
#' @param ... Ignored
#' @returns Returns a named `list` with account details.
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_get_account()
#' }
#' @include helpers.R
#' @family authentication-functions
#' @export
cds_get_account <- function(token = cds_get_token(), ...) {
  .base_url |>
    paste0("/profiles/v1/account/", sep = "") |>
    .execute_request(token)
}

#' Get Prometheus metrics for account
#' 
#' Obtain account metrics that can be interpreted with [Prometheus](https://prometheus.io/)
#' @inheritParams cds_check_authentication
#' @param ... Ignored
#' @returns Returns text that can be interpreted with [Prometheus](https://prometheus.io/)
#' @examples
#' if (interactive() && cds_token_works()) {
#'   cds_account_metrics()
#' }
#' @include helpers.R
#' @family authentication-functions
#' @export
cds_account_metrics <- function(token = cds_get_token(), ...) {
  .base_url |>
    paste0("/profiles/v1/metrics/", sep = "") |>
    .make_request(token) |>
    httr2::req_error(body = .req_error) |>
    httr2::req_perform() |>
    httr2::resp_body_string()
}

test_that("Authentication works", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_no_error({
    cds_check_authentication()
  })
})

test_that("Getting account details works", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_true({
    cds_get_account()$origin_portal[[1]] == "c3s"
  })
})

test_that("Can get account metrics", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not(cds_token_works())
  expect_no_error({
    cds_account_metrics()
  })
})

test_that("Token can be set", {
  expect_true({
    my_token <- cds_get_token()
    cds_set_token("foobar", "option")
    cds_set_token("foobar", "sysenv")
    result <- getOption("CDSAPI_KEY") == "foobar" &&
      Sys.getenv("CDSAPI_KEY") == "foobar"
    options(CDSAPI_KEY = NULL)
    cds_set_token(my_token, "sysenv")
    result
  })
})

.base_url <- "https://cds.climate.copernicus.eu/api"

.make_request <- function(x, token = "", method = "GET") {
  x |>
    httr2::request() |>
    httr2::req_headers(
      `PRIVATE-TOKEN` = token,
      `User-Agent` = "r_package") |>
    httr2::req_method(method) |>
    httr2::req_retry(max_tries = 3, retry_on_failure = TRUE, after = \(x) NA)
}

.execute_request <- function(x, token = "", method = "GET", req_body = NULL) {
  x |>
    .make_request(token, method) |>
    httr2::req_body_json(req_body, auto_unbox = TRUE) |>
    httr2::req_error(body = .req_error) |>
    httr2::req_perform() |>
    httr2::resp_body_json()
}

.req_error <- function(body) {
  ct <- httr2::resp_content_type(body)
  body <-
    switch(
    ct,
    `application/json` = {
      body <- httr2::resp_body_json(body)
      if (is.list(body$detail)) body$detail <- body$detail[[1]]$msg
      if (!is.null(body$detail) && nchar(body$detail) > 3) {
        body$detail <- tryCatch({
          result <-
            substr(body$detail, 2, nchar(body$detail) - 1) |>
            stringr::str_replace_all("'", "\"") |>
            stringr::str_replace_all("\\(", "[") |>
            stringr::str_replace_all("\\)", "]") |>
            jsonlite::fromJSON()
          result$msg
        }, error = function(e) body$detail)
        body
      } else {
        body$detail <- NULL
        body
      }
    },
    `text/html` = {
      list(detail = httr2::resp_body_string(body))
    },
    list(detail = "")
  )
  c(body$title, body$detail)
}

.simplify <- function(x) {
  x <- lapply(x, tibble::enframe) |>
    lapply(tidyr::pivot_wider, names_from = "name", values_from = "value") |>
    dplyr:: bind_rows()
  unnest_col <-
    lapply(x, \(x) lengths(x) == 1) |>
    lapply(any) |>
    unlist()
  tidyr::unnest(x, names(unnest_col)[unnest_col])
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tryCatch({
  Sys.setlocale("LC_ALL", "English")
})
library(ggplot2)
theme_set(theme_light())
dataset_form <- NULL
request <- NULL
job <- NULL
file_result <- NULL
filename <- "skip"

## ----token--------------------------------------------------------------------
library(CopernicusClimate)

message(
  "The machine that rendered this vignette ",
  ifelse(
    cds_token_works(), "has", "does not have"),
  " a working token")


## ----get-licence, message=FALSE-----------------------------------------------
library(dplyr)

tryCatch({
  licence_info <-
    cds_dataset_form("reanalysis-era5-pressure-levels") |>
      filter(name == "licences")

  licence_info <- licence_info$details[[1]]$details$licences[[1]]
  print(licence_info)
}, error = \(e) message("Failed to get licence info"))

## ----listing------------------------------------------------------------------
tryCatch({
  cds_list_datasets()
}, error = \(e) message("Failed to retrieve datasets"))

## ----search-------------------------------------------------------------------
tryCatch({
  cds_search_datasets(search = "rain", keywords = "Temporal coverage: Future")
}, error = \(e) message("Failed to search for datasets"))

## ----dataset-form-------------------------------------------------------------
tryCatch({
  dataset_form <-
    cds_dataset_form("reanalysis-era5-pressure-levels")
  dataset_form
}, error = \(e) message("Failed to retrieve request form"))

## ----possible-values, eval=!is.null(dataset_form)-----------------------------
values <-
  dataset_form |>
  filter(name == "pressure_level") |>
  pull("details")

values[[1]]$details$values |> unlist()

## ----full-request-------------------------------------------------------------
tryCatch({
  request <- cds_build_request("reanalysis-era5-pressure-levels")
  summary(request)
}, error = \(e) message("Failed to build request"))

## ----specific-request---------------------------------------------------------
tryCatch({
  request <- cds_build_request(
    "reanalysis-era5-pressure-levels",
    variable       = "temperature",
    pressure_level = "1000",
    year           = "2025",
    month          = "01",
    day            = "01",
    area           = c(n = 60, w = -5, s = 40, e = 10),
    data_format    = "netcdf")
  summary(request)
}, error = \(e) message("Failed to build the request"))

## ----estimate-full------------------------------------------------------------
if (cds_token_works()) {
  tryCatch({
    cds_estimate_costs("reanalysis-era5-pressure-levels")
  }, error = \(e) message("Failed to estimate costs"))
} else {
  message("You need a working token to estimate costs")
}

## ----estimate-detailed, eval=!is.null(request)--------------------------------
if (cds_token_works()) {
  tryCatch({
    cds_estimate_costs(request)
  }, error = \(e) message("Failed to estimate costs"))
} else {
  message("You need a working token to estimate costs")
}

## ----submit, message=FALSE----------------------------------------------------
if (cds_token_works()) {
  tryCatch({
    job <-
      cds_submit_job(request)
    job
  }, error = \(e) message("Failed to submit job"))
} else {
  message("You need a working token to submit a request")
}

## ----job-status, eval=!is.null(job)-------------------------------------------
if (cds_token_works()) {
  cds_list_jobs(job$jobID)
} else {
  message("You need a working token to get a job status")
}

## ----download, message=FALSE, eval=!is.null(job)------------------------------
filename <- "result.nc"
if (cds_token_works()) {
  tryCatch({
    file_result <- cds_download_jobs(job$jobID, tempdir(), filename)
  }, error = \(e) message("Failed to download job"))
} else {
  message("Downloading data only works with a valid token")
}

## ----plot, fig.width=7, fig.height=3, message = FALSE, fig.alt="Plot created from downloaded data"----
fn <- file.path(tempdir(), filename)

if (file.exists(fn)) {
  
  library(stars)
  library(ggplot2)
  
  result <- read_mdim(fn)
  
  ggplot() +
    geom_stars(data = result) +
    coord_sf() +
    facet_wrap(~strftime(as.POSIXct(valid_time), "%H:%M"), nrow = 3) +
    scale_fill_viridis_c(option = "turbo") +
    labs(x = NULL, y = NULL, fill = "Temperature [K]") +
    theme(axis.text = element_blank())

} else {
  message("File wasn't downloaded")
}

#' Cite a dataset
#' 
#' Use this function to obtain citation details for a specific dataset
#' @param dataset The name of a dataset to be cited.
#' @param ... Ignored
#' @returns Returns a `BibEntry`-class object, with citation details for the requested
#' dataset
#' @examples
#' if (interactive()) {
#'   cds_cite_dataset("reanalysis-era5-pressure-levels")
#' }
#' @family helper-functions
#' @export
cds_cite_dataset <- function(dataset, ...) {
  if (requireNamespace("RefManageR")) {
    ds <- cds_list_datasets(dataset)
    RefManageR::GetBibEntryWithDOI(ds$`sci:doi`)
  } else {
    rlang::abort(c(x = "This function needs package RefManageR",
                   i = "Please install and try again"))
  }
}

test_that("Dataset can be cited", {
  skip_on_cran()
  skip_if_offline()
  skip_if_not_installed("RefManageR")
  testthat::expect_s3_class({
    cds_cite_dataset("reanalysis-era5-pressure-levels") |>
      suppressMessages()
  }, "BibEntry")
})

#' Translate Python Climate Data Store request to R request
#' 
#' When looking for a dataset on <https://cds.climate.copernicus.eu/datasets>,
#' you have the option to copy the API request code to the clipboard. However,
#' this is Python code and cannot be used directly in this package.
#' Use this function to translate the code to a request that can be handled by
#' the package. For more details see `vignette("translate")`
#' @param text A `character` string containing the Python code copied
#' from a dataset download website <https://cds.climate.copernicus.eu/datasets>.
#' When missing, the function will use any text on the system clipboard.
#' This means that you can copy the API request code from the website to the
#' clipboard, then call this function without arguments.
#' @param ... Ignored
#' @returns A named `list` that can be used as input for the functions
#' `cds_submit_job()` and `cds_estimate_costs()`
#' @examples
#' python_code <-
#' "import cdsapi
#' dataset = \"reanalysis-era5-land\"
#' request = {
#'   \"variable\": [\"2m_dewpoint_temperature\"],
#'   \"year\": \"2025\",
#'   \"month\": \"01\",
#'   \"day\": [\"01\"],
#'   \"time\": [
#'     \"00:00\", \"01:00\", \"02:00\",
#'     \"03:00\", \"04:00\", \"05:00\",
#'     \"06:00\", \"07:00\", \"08:00\",
#'     \"09:00\", \"10:00\", \"11:00\",
#'     \"12:00\", \"13:00\", \"14:00\",
#'     \"15:00\", \"16:00\", \"17:00\",
#'     \"18:00\", \"19:00\", \"20:00\",
#'     \"21:00\", \"22:00\", \"23:00\"
#'   ],
#'   \"data_format\": \"netcdf\",
#'   \"download_format\": \"unarchived\"
#' }
#' 
#' client = cdsapi.Client()
#' client.retrieve(dataset, request).download()
#' "
#' 
#' cds_python_to_r(python_code)
#' @family helper-functions
#' @export
cds_python_to_r <- function(text, ...) {
  if (missing(text)) {
    if (requireNamespace("clipr")) {
      text <- clipr::read_clip() |>
        suppressWarnings()
      if (is.null(text)) rlang::abort(c(
        x = "System clipboard contained no readable text",
        i = "Pass argument 'text' to the function"
      ))
    } else {
      if (clipr::clipr_available()) {
        rlang::abort(
          c(x = "Trying to get 'text', however, system clipboard is not accessible",
            i = "Pass the Python code explicitly as 'text' argument")
        )
      } else {
        rlang::abort(
          c(x = "Trying to get 'text' from clipboard, but missing required 'clipr' package",
            i = "Install package 'clipr' and try again")
        )
      }
    }
  }
  ## collapse + strip spaces and tabs:
  text <- gsub("[ \t]", "", paste(text, collapse = "\n"))
  tryCatch({
    ## extract dataset from text:
    dataset <-
      text |>
      stringr::str_extract("dataset=\".*\"") |>
      stringr::str_replace_all("(\"|dataset=)", "")
    
    ## extract request args from text:
    req <-
      text |>
      stringr::str_extract("request=\\{(.|\n)*\\}") |>
      stringr::str_replace("^request=", "") |>
      jsonlite::fromJSON(simplifyVector = FALSE)
    attributes(req)$dataset <- dataset
    req
  }, error = function(e) rlang::abort(
    c(x = "Failed to convert text to CDSAPI request",
      i = "Check the input text (also the clipboard if relevant)")))
}

test_that("Source code should not have things on TODO list", {
  skip_on_ci()
  skip_on_cran()
  expect_false({
    skip_if(length(unclass(packageVersion("CopernicusClimate"))[[1]]) > 3,
            "Skipping during development")
    r_path <- normalizePath(test_path()) |> dirname() |> dirname()
    files_to_check <- list.files(r_path, pattern = ".[r|R]$|NEWS|DESCRIPTION|README\\.md",
                                 recursive = TRUE, full.names = TRUE)
    files_to_check <- files_to_check[!endsWith(files_to_check, "test-dev.r")]
    any(
      unlist(
        lapply(files_to_check, function(file) {
          content <- suppressWarnings(readLines(file))
          result  <- grepl("TODO", content) & !grepl("grepl\\(\"TODO\"", content) & !grepl("on TODO list", content)
          if (any(result)) {
            warning(sprintf("File `%s` has items on TODO list at lines `%s`",
                            file, paste(which(result), collapse = "`, `")))
          }
          any(result)
        })
      )
    )
  })
})

---
title: "Downloading from Copernicus Climate Service"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Downloading from Copernicus Climate Service}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tryCatch({
  Sys.setlocale("LC_ALL", "English")
})
library(ggplot2)
theme_set(theme_light())
```

## Introduction

When obtaining data from Copernicus Climate Data Service you cannot download
the data directly. You need to know which data you want, submit a request for
a specific dataset. Wait for your request to complete, and if successful, download
the data. The `CopernicusClimate` package has functions to facilitate this process.
This vignette will walk you through the different steps to download data.

 * [Finding datasets](#finding-datasets)
 * [Requesting a dataset](#specifying-a-request)
 * [Tracking submitted requests](#tracking-submitted-requests)
 * [Download data](#downloading-data)

But before you can even get started, there are some things you have to prepare first,
as explained in the following section.

## Prerequisites

### Access token

This R package is built around the Application Programming Interface (API) provided
by C3S. Many of the features of this API require to identify yourself, for which
a 'key' or API token is used. You can get one by creating an account at
<https://cds.climate.copernicus.eu/profile>.

Once you have an account you can generate (or refresh) an API key. You can use this
token by means of the `token` argument in many of the functions of this package. But
rather then providing the key separately each time, you can use the key throughout
your R session by setting it once with `cds_set_token()`.

However, if you want to share your work, it is not very secure to keep your strictly
personal key hard coded in your script. Furthermore, setting the key with `cds_set_token()`
does not persist across sessions. Instead, you could set it as an option in your
`.rprofile` file, or as a environmental variable on your system. In both cases the
variable should be named `CDSAPI_KEY`. This variable is automatically picked up by
`cds_get_token()` and you don't have to specify it anywhere in your script.

You can check if your token works with `cds_token_works()`:

```{r token}
library(CopernicusClimate)

message(
  "The machine that rendered this vignette ",
  ifelse(
    cds_token_works(), "has", "does not have"),
  " a working token")

```

### Licences

In order to download datasets you need to accept its accompanying licence.
You can use `cds_dataset_form()` to inspect under which licence a dataset
is provided, like so:

```{r get-licence, message=FALSE}
library(dplyr)

licence_info <-
  cds_dataset_form("reanalysis-era5-pressure-levels") |>
    filter(name == "licences")

licence_info <- licence_info$details[[1]]$details$licences[[1]]
print(licence_info)
```

You can accept this licence by calling
`cds_accept_licence(licence_info$id, licence_info$revision)`. You only need to do
this once for every licence. Accepted licences are stored with your account and
can be listed with `cds_accepted_licences()`. Without accepting required licences
you cannot submit a successful request for downloading it.

## Finding datasets

### Websites

If you want a visual interface for exploring available datasets, you can use
your web browser and visit either the
[Climate Data Store](https://cds.climate.copernicus.eu/datasets) or
[STAC catalogue](https://cds.climate.copernicus.eu/stac-browser/).
Both allow you to navigate through the treasures of information, and identify which
dataset best serves your needs. You can even copy the request code
to the system clipboard and use this package to get the request from the clipboard
and execute it. This is illustrated in `vignette("translate")`.

### Programmatically

You can also use this R package to look for datasets. You could start by listing them
all:

```{r listing}
cds_list_datasets()
```

But you can also look for specific datasets using free search text and / or predefined
keywords:

```{r search}
cds_search_datasets(search = "rain", keywords = "Temporal coverage: Future")
```

Use `cds_catalogue_vocabulary()` to list available predefined keywords.

You will see that either approach results in a `data.frame` with a  column named `id`.
You can use this `id` to refer to when setting up a request for download.

### Favourite datasets

You can also mark your favourite datasets with a star using `cds_assign_star()`. You
can get list your favourite datasets with `cds_starred()`. This makes it easier to
find datasets you use a lot. You can remove a star with `cds_remove_star()`.

## Specifying a request

In many cases you cannot download an entire dataset at once, because it it
too large. This means you have to specify a subset that you want to have.

### What are my options?

How do you know what options you have to subset a dataset? These options differ
for each dataset, so there is no straightforward answer. However, you can inspect
what options you have for a specific dataset. You can start by obtaining the
`cds_dataset_form()`.

```{r dataset-form}
dataset_form <-
  cds_dataset_form("reanalysis-era5-pressure-levels")

dataset_form
```

This results in a `data.frame` listing which aspects of a dataset you can
select from. Each row represents an aspect (except for the row with the `name`
`"licences"`). The column `details` contains information about the available
values. You could for instance look at the possible values for the `pressure_level`:

```{r possible-values}
values <-
  dataset_form |>
  filter(name == "pressure_level") |>
  pull("details")

values[[1]]$details$values |> unlist()
```

Using this information you can start building your request using `cds_build_request()`.
You can start by just specifying your dataset:

```{r full-request}
request <- cds_build_request("reanalysis-era5-pressure-levels")
summary(request)
```

The function `cds_build_request()` will automatically add all required parameters
to the request and fills it with either their default value, if available, or all allowed
values otherwise. The request built above will ask for the complete dataset
in the default product type, plus data and download format. As I will explain in the
following section, this request will fail for most users. So let's narrow it down:

```{r specific-request}
request <- cds_build_request(
  "reanalysis-era5-pressure-levels",
  variable       = "temperature",
  pressure_level = "1000",
  year           = "2025",
  month          = "01",
  day            = "01",
  area           = c(n = 60, w = -5, s = 40, e = 10),
  data_format    = "netcdf")
summary(request)
```

This looks like a reasonable request.

### How much can I get?

As mentioned before, the amount of data that can be requested for each
download is restricted. In order to test how much a request would cost you
can call `cds_estimate_costs()`. Using the example above, if you want to
download the full dataset, the estimated costs are as follows:

```{r estimate-full}
if (cds_token_works()) {
  cds_estimate_costs("reanalysis-era5-pressure-levels")
} else {
  message("You need a working token to estimate costs")
}
```

In this example the costs exceed the limit, such that this request
will fail. If we estimate the costs for the more restricted request,
we get:

```{r estimate-detailed}
if (cds_token_works()) {
  cds_estimate_costs(request)
} else {
  message("You need a working token to estimate costs")
}
```

This is a request that we can afford.

## Submitting a request

Once you have established which dataset you want to download and how you wish to subset
it, you can submit a request to C3S. Let's submit the request as shown above:

```{r submit, message=FALSE}
if (cds_token_works()) {
  job <-
    cds_submit_job(request)
  job
} else {
  message("You need a working token to submit a request")
}
```

By default this function will wait until the request has been processed by
C3S. But when you set the argument `wait = FALSE`, the function will
return immediately. In that case, you can submit multiple jobs where you
don't have to wait for each individual request to complete.

## Tracking submitted requests

When submitting a request and choose not to wait for it to complete, you
may want to track the progress of your request. You can use `cds_list_jobs()`
to list all your submitted jobs. If you want the status of a specific job,
you can use its identifier (id). You were sent this id when you submitted it earlier.
So we can have a look at the status of our job submitted above:

```{r job-status}
if (cds_token_works()) {
  cds_list_jobs(job$jobID)
} else {
  message("You need a working token to get a job status")
}
```

## Downloading data

Now that we have submitted the request we can download it (if it is completed
successfully) with `cds_download_jobs()`. If you don't specify a job identifier,
it will download all (previously submitted) successful jobs. You can also
download one or more specific jobs. Note that this function will use parallel downloads
which should give you some performance advantage when downloading multiple jobs.
For now let's try to download the submitted job:

```{r download, message=FALSE}
filename <- "result.nc"
if (cds_token_works()) {
  file_result <- cds_download_jobs(job$jobID, tempdir(), filename)
} else {
  message("Downloading data only works with a valid token")
}
```

Now you can do whatever it is you want to do with the data:

```{r plot, fig.width=7, fig.height=3, message = FALSE, fig.alt="Plot created from downloaded data"}
fn <- file.path(tempdir(), filename)

if (file.exists(fn)) {
  
  library(stars)
  library(ggplot2)
  
  result <- read_mdim(fn)
  
  ggplot() +
    geom_stars(data = result) +
    coord_sf() +
    facet_wrap(~strftime(valid_time, "%H:%M"), nrow = 3) +
    scale_fill_viridis_c(option = "turbo") +
    labs(x = NULL, y = NULL, fill = "Temperature [K]") +
    theme(axis.text = element_blank())

} else {
  message("File wasn't downloaded")
}
```

---
title: "Translate API Code"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Translate API Code}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Locating a dataset

You could use the R functions `cds_search_datasets()` and `cds_list_datasets()` to
look for [Copernicus Climate Change Service (C3S)](https://climate.copernicus.eu/) data.
If you prefer using a graphical user interface, you can use the website to find
a suitable dataset: <https://cds.climate.copernicus.eu/datasets>.

The screen recording below shows how to look for a dataset on the website, and then
copy the request code for that dataset to the clipboard. Note this recording is
only showing the essentials, and no specific subsetting selections are made, causing
the actual request to be incomplete. But the principle remains the same.

![Demonstration of copying request code to clipboard](https://raw.githubusercontent.com/pepijn-devries/CopernicusClimate/refs/heads/main/pkgdown/media/copy_api_code.gif)

## Downloading the dataset

Once the request code is on your system's clipboard, you can go to your R console
and download the dataset, by first submitting the request with `cds_submit_job()`.
By omitting the dataset argument, the function will automatically look for a request
on the clipboard. After the request has been processed, you can download the dataset
with `cds_download_jobs()`. All of this is illustrated in the screen recording below.

![Demonstration of downloading using a request on the clipboard](https://raw.githubusercontent.com/pepijn-devries/CopernicusClimate/refs/heads/main/pkgdown/media/download_from_clipboard.gif)

You can also arrange your workflow completely in R, without using a web browser.
For more details on this workflow check out `vignette("download")`.

## Translation

Under the hood of all this is the function `cds_python_to_r()`. It is used to translate
request code from the website to a request that can be handled by this R package.
In essence it extracts all relevant information from the Python codes and turns it into
a named list.

---
title: "Downloading from Copernicus Climate Service"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Downloading from Copernicus Climate Service}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
tryCatch({
  Sys.setlocale("LC_ALL", "English")
})
library(ggplot2)
theme_set(theme_light())
dataset_form <- NULL
request <- NULL
job <- NULL
file_result <- NULL
filename <- "skip"
```

## Introduction

When obtaining data from Copernicus Climate Data Service you cannot download
the data directly. You need to know which data you want, submit a request for
a specific dataset. Wait for your request to complete, and if successful, download
the data. The `CopernicusClimate` package has functions to facilitate this process.
This vignette will walk you through the different steps to download data.

 * [Finding datasets](#finding-datasets)
 * [Requesting a dataset](#specifying-a-request)
 * [Tracking submitted requests](#tracking-submitted-requests)
 * [Download data](#downloading-data)

But before you can even get started, there are some things you have to prepare first,
as explained in the following section.

## Prerequisites

### Access token

This R package is built around the Application Programming Interface (API) provided
by C3S. Many of the features of this API require to identify yourself, for which
a 'key' or API token is used. You can get one by creating an account at
<https://cds.climate.copernicus.eu/profile>.

Once you have an account you can generate (or refresh) an API key. You can use this
token by means of the `token` argument in many of the functions of this package. But
rather then providing the key separately each time, you can use the key throughout
your R session by setting it once with `cds_set_token()`.

However, if you want to share your work, it is not very secure to keep your strictly
personal key hard coded in your script. Furthermore, setting the key with `cds_set_token()`
does not persist across sessions. Instead, you could set it as an option in your
`.rprofile` file, or as a environmental variable on your system. In both cases the
variable should be named `CDSAPI_KEY`. This variable is automatically picked up by
`cds_get_token()` and you don't have to specify it anywhere in your script.

You can check if your token works with `cds_token_works()`:

```{r token}
library(CopernicusClimate)

message(
  "The machine that rendered this vignette ",
  ifelse(
    cds_token_works(), "has", "does not have"),
  " a working token")

```

### Licences

In order to download datasets you need to accept its accompanying licence.
You can use `cds_dataset_form()` to inspect under which licence a dataset
is provided, like so:

```{r get-licence, message=FALSE}
library(dplyr)

tryCatch({
  licence_info <-
    cds_dataset_form("reanalysis-era5-pressure-levels") |>
      filter(name == "licences")

  licence_info <- licence_info$details[[1]]$details$licences[[1]]
  print(licence_info)
}, error = \(e) message("Failed to get licence info"))
```

You can accept this licence by calling
`cds_accept_licence(licence_info$id, licence_info$revision)`. You only need to do
this once for every licence. Accepted licences are stored with your account and
can be listed with `cds_accepted_licences()`. Without accepting required licences
you cannot submit a successful request for downloading it.

## Finding datasets

### Websites

If you want a visual interface for exploring available datasets, you can use
your web browser and visit either the
[Climate Data Store](https://cds.climate.copernicus.eu/datasets) or
[STAC catalogue](https://cds.climate.copernicus.eu/stac-browser/).
Both allow you to navigate through the treasures of information, and identify which
dataset best serves your needs. You can even copy the request code
to the system clipboard and use this package to get the request from the clipboard
and execute it. This is illustrated in `vignette("translate")`.

### Programmatically

You can also use this R package to look for datasets. You could start by listing them
all:

```{r listing}
tryCatch({
  cds_list_datasets()
}, error = \(e) message("Failed to retrieve datasets"))
```

But you can also look for specific datasets using free search text and / or predefined
keywords:

```{r search}
tryCatch({
  cds_search_datasets(search = "rain", keywords = "Temporal coverage: Future")
}, error = \(e) message("Failed to search for datasets"))
```

Use `cds_catalogue_vocabulary()` to list available predefined keywords.

You will see that either approach results in a `data.frame` with a  column named `id`.
You can use this `id` to refer to when setting up a request for download.

### Favourite datasets

You can also mark your favourite datasets with a star using `cds_assign_star()`. You
can get list your favourite datasets with `cds_starred()`. This makes it easier to
find datasets you use a lot. You can remove a star with `cds_remove_star()`.

## Specifying a request

In many cases you cannot download an entire dataset at once, because it it
too large. This means you have to specify a subset that you want to have.

### What are my options?

How do you know what options you have to subset a dataset? These options differ
for each dataset, so there is no straightforward answer. However, you can inspect
what options you have for a specific dataset. You can start by obtaining the
`cds_dataset_form()`.

```{r dataset-form}
tryCatch({
  dataset_form <-
    cds_dataset_form("reanalysis-era5-pressure-levels")
  dataset_form
}, error = \(e) message("Failed to retrieve request form"))
```

This results in a `data.frame` listing which aspects of a dataset you can
select from. Each row represents an aspect (except for the row with the `name`
`"licences"`). The column `details` contains information about the available
values. You could for instance look at the possible values for the `pressure_level`:

```{r possible-values, eval=!is.null(dataset_form)}
values <-
  dataset_form |>
  filter(name == "pressure_level") |>
  pull("details")

values[[1]]$details$values |> unlist()
```

Using this information you can start building your request using `cds_build_request()`.
You can start by just specifying your dataset:

```{r full-request}
tryCatch({
  request <- cds_build_request("reanalysis-era5-pressure-levels")
  summary(request)
}, error = \(e) message("Failed to build request"))
```

The function `cds_build_request()` will automatically add all required parameters
to the request and fills it with either their default value, if available, or all allowed
values otherwise. The request built above will ask for the complete dataset
in the default product type, plus data and download format. As I will explain in the
following section, this request will fail for most users. So let's narrow it down:

```{r specific-request}
tryCatch({
  request <- cds_build_request(
    "reanalysis-era5-pressure-levels",
    variable       = "temperature",
    pressure_level = "1000",
    year           = "2025",
    month          = "01",
    day            = "01",
    area           = c(n = 60, w = -5, s = 40, e = 10),
    data_format    = "netcdf")
  summary(request)
}, error = \(e) message("Failed to build the request"))
```

This looks like a reasonable request.

### How much can I get?

As mentioned before, the amount of data that can be requested for each
download is restricted. In order to test how much a request would cost you
can call `cds_estimate_costs()`. Using the example above, if you want to
download the full dataset, the estimated costs are as follows:

```{r estimate-full}
if (cds_token_works()) {
  tryCatch({
    cds_estimate_costs("reanalysis-era5-pressure-levels")
  }, error = \(e) message("Failed to estimate costs"))
} else {
  message("You need a working token to estimate costs")
}
```

In this example the costs exceed the limit, such that this request
will fail. If we estimate the costs for the more restricted request,
we get:

```{r estimate-detailed, eval=!is.null(request)}
if (cds_token_works()) {
  tryCatch({
    cds_estimate_costs(request)
  }, error = \(e) message("Failed to estimate costs"))
} else {
  message("You need a working token to estimate costs")
}
```

This is a request that we can afford.

## Submitting a request

Once you have established which dataset you want to download and how you wish to subset
it, you can submit a request to C3S. Let's submit the request as shown above:

```{r submit, message=FALSE}
if (cds_token_works()) {
  tryCatch({
    job <-
      cds_submit_job(request)
    job
  }, error = \(e) message("Failed to submit job"))
} else {
  message("You need a working token to submit a request")
}
```

By default this function will wait until the request has been processed by
C3S. But when you set the argument `wait = FALSE`, the function will
return immediately. In that case, you can submit multiple jobs where you
don't have to wait for each individual request to complete.

## Tracking submitted requests

When submitting a request and choose not to wait for it to complete, you
may want to track the progress of your request. You can use `cds_list_jobs()`
to list all your submitted jobs. If you want the status of a specific job,
you can use its identifier (id). You were sent this id when you submitted it earlier.
So we can have a look at the status of our job submitted above:

```{r job-status, eval=!is.null(job)}
if (cds_token_works()) {
  cds_list_jobs(job$jobID)
} else {
  message("You need a working token to get a job status")
}
```

## Downloading data

Now that we have submitted the request we can download it (if it is completed
successfully) with `cds_download_jobs()`. If you don't specify a job identifier,
it will download all (previously submitted) successful jobs. You can also
download one or more specific jobs. Note that this function will use parallel downloads
which should give you some performance advantage when downloading multiple jobs.
For now let's try to download the submitted job:

```{r download, message=FALSE, eval=!is.null(job)}
filename <- "result.nc"
if (cds_token_works()) {
  tryCatch({
    file_result <- cds_download_jobs(job$jobID, tempdir(), filename)
  }, error = \(e) message("Failed to download job"))
} else {
  message("Downloading data only works with a valid token")
}
```

Now you can do whatever it is you want to do with the data:

```{r plot, fig.width=7, fig.height=3, message = FALSE, fig.alt="Plot created from downloaded data"}
fn <- file.path(tempdir(), filename)

if (file.exists(fn)) {
  
  library(stars)
  library(ggplot2)
  
  result <- read_mdim(fn)
  
  ggplot() +
    geom_stars(data = result) +
    coord_sf() +
    facet_wrap(~strftime(as.POSIXct(valid_time), "%H:%M"), nrow = 3) +
    scale_fill_viridis_c(option = "turbo") +
    labs(x = NULL, y = NULL, fill = "Temperature [K]") +
    theme(axis.text = element_blank())

} else {
  message("File wasn't downloaded")
}
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieve.R
\name{cds_download_jobs}
\alias{cds_download_jobs}
\title{Download specific jobs}
\usage{
cds_download_jobs(job_id, destination, names, ..., token = cds_get_token())
}
\arguments{
\item{job_id}{If a specific job identifier is listed here, only the files resulting
from those jobs are downloaded. If left blank, all successful jobs are downloaded.}

\item{destination}{Destination path to store downloaded files.}

\item{names}{File names for the downloaded files. If missing, the cryptic hexadecimal
file name is taken from the job.}

\item{...}{Ignored}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
A \code{data.frame} of all downloaded files. Contains a column \code{local} with the path
to the locally stored files.
}
\description{
After submitting one or more jobs with \code{cds_submit_job()}, you can download the resulting
files with \code{cds_download_jobs()}. See \code{vignette("download")} for more details.
}
\examples{
if (interactive() && cds_token_works()) {
  job <- cds_submit_job(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = c(n = 55, w = -1, s = 50, e = 10),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf"
    )
  cds_download_jobs(job$jobID, tempdir())
}
}
\seealso{
Other job-functions:
\code{\link[=cds_delete_job]{cds_delete_job()}},
\code{\link[=cds_job_results]{cds_job_results()}},
\code{\link[=cds_list_jobs]{cds_list_jobs()}},
\code{\link[=cds_submit_job]{cds_submit_job()}}
}
\concept{job-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieve.R
\name{cds_submit_job}
\alias{cds_submit_job}
\title{Submit a download job for a dataset}
\usage{
cds_submit_job(
  dataset,
  ...,
  wait = TRUE,
  check_quota = TRUE,
  check_licence = TRUE,
  token = cds_get_token()
)
}
\arguments{
\item{dataset}{The dataset name to be downloaded, or a \code{list} returned by
\code{cds_build_request()}. When this argument is missing the function will
attempt to build a request with text on the clipboard using \code{cds_python_to_r()}.}

\item{...}{Subsetting parameters passed onto \code{cds_build_request()}. Should
be empty when \code{dataset} is missing or a \code{list}.}

\item{wait}{A \code{logical} value indicating if the function should wait for the
submitted job to finish. Set it to \code{FALSE} to continue without waiting}

\item{check_quota}{Each account has a quota of data that can be downloaded.
If this argument is set to \code{TRUE} (default) it is checked if the request doesn't
exceed your quota. Set it to \code{FALSE} to skip this step and speed up the submission.}

\item{check_licence}{Datasets generally require you to accept certain terms of use.
If this argument is set to \code{TRUE} (default), it will be checked if you have accepted all
required licences for the submitted request. Set it to \code{FALSE} to skip this step and
speed up the submission.}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
Returns a \code{data.frame} containing information about the submitted job.
}
\description{
Submit a request to the Copernicus Climate Data Service to download
(part of) a dataset. If the request is successful, a job identifier is
returned which can be used to actually download the data (\code{cds_download_jobs()}).
}
\examples{
if (interactive() && cds_token_works()) {
  job <- cds_submit_job(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = c(n = 55, w = -1, s = 50, e = 10),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf"
    )

  ## Or split into two separate steps:
  
  req <- cds_build_request(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = c(n = 55, w = -1, s = 50, e = 10),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf"
    )
  job <- cds_submit_job(req)
}
}
\seealso{
Other job-functions:
\code{\link[=cds_delete_job]{cds_delete_job()}},
\code{\link[=cds_download_jobs]{cds_download_jobs()}},
\code{\link[=cds_job_results]{cds_job_results()}},
\code{\link[=cds_list_jobs]{cds_list_jobs()}}
}
\concept{job-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/licenses.R
\name{cds_accept_licence}
\alias{cds_accept_licence}
\title{Accept a specific licence}
\usage{
cds_accept_licence(license, revision, ..., token = cds_get_token())
}
\arguments{
\item{license}{The license id you wish to accept.}

\item{revision}{The revision number of the license you are accepting.
Should always be the latest revision.}

\item{...}{Ignored}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
Returns a \code{data.frame} containing the accepted license
}
\description{
Call this function if you wish to accept a specific license.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_accept_licence("cc-by", 1)
}
}
\seealso{
Other licences:
\code{\link[=cds_accepted_licences]{cds_accepted_licences()}},
\code{\link[=cds_list_licences]{cds_list_licences()}}
}
\concept{licences}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/licenses.R
\name{cds_accepted_licences}
\alias{cds_accepted_licences}
\title{List accepted licences}
\usage{
cds_accepted_licences(
  scope = c("all", "dataset", "portal"),
  ...,
  token = cds_get_token()
)
}
\arguments{
\item{scope}{Scope of the licenses to be listed should be one of \code{"all"} (default),
\code{"dataset"} or \code{"portal"}.}

\item{...}{Ignored}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
Returns a \code{data.frame} listing the accepted licenses.
}
\description{
In order to use specific features of the Climate Data Service, you may first
need to accept specific licences. This function will list all licenses you
have accepted.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_accepted_licences("portal")
}
}
\seealso{
Other licences:
\code{\link[=cds_accept_licence]{cds_accept_licence()}},
\code{\link[=cds_list_licences]{cds_list_licences()}}
}
\concept{licences}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalogue.R
\name{cds_dataset_form}
\alias{cds_dataset_form}
\title{Obtain an overview of options to subset a dataset}
\usage{
cds_dataset_form(dataset, ...)
}
\arguments{
\item{dataset}{A name of a dataset to explore}

\item{...}{Ignored}
}
\value{
A \code{data.frame} with aspects of the dataset that can be subsetted, when defining
a job. See also \code{cds_submit_job()} and \code{cds_build_request()}
}
\description{
This function provides an overview of parameters that can be used to subset a
dataset. It can help you set up a request with \code{cds_submit_job()} or \code{cds_build_request()}.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_dataset_form("reanalysis-era5-pressure-levels")
}
}
\seealso{
Other helper-functions:
\code{\link[=cds_build_request]{cds_build_request()}},
\code{\link[=cds_cite_dataset]{cds_cite_dataset()}},
\code{\link[=cds_estimate_costs]{cds_estimate_costs()}},
\code{\link[=cds_python_to_r]{cds_python_to_r()}}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalogue.R
\name{cds_list_datasets}
\alias{cds_list_datasets}
\alias{cds_search_datasets}
\title{List or search Climate Data Service datasets}
\usage{
cds_list_datasets(dataset, ...)

cds_search_datasets(search, keywords, page = 0, limit = 50, ...)
}
\arguments{
\item{dataset}{A specific dataset for which to list details. If missing
all datasets are listed.}

\item{...}{Ignored}

\item{search}{A string containing free text search terms to look for in the available datasets.}

\item{keywords}{A (vector of) string containing specific keywords. Should be listed in
\code{cds_catalogue_vocabulary()}}

\item{page}{When there are more search results than \code{limit}, results are paginated.
Use \code{page} to specify which page to return starting at \code{0} (default).}

\item{limit}{Use to limit the number of search results. Defaults to \code{50}.}
}
\value{
A \code{data.frame} listing the datasets from the Climate Data Service. In
case of \code{cds_search_datasets()}, attributes named \code{SearchMeta} are added containing
the number of matching and number of returned datasets.
}
\description{
This will help you decide which datasets you wish to obtain.
}
\examples{
if (interactive()) {
  cds_list_datasets()
  cds_search_datasets("rain", "Temporal coverage: Future")
}
}
\seealso{
Other exploration-functions:
\code{\link[=cds_catalogue_vocabulary]{cds_catalogue_vocabulary()}},
\code{\link[=cds_starred]{cds_starred()}}
}
\concept{exploration-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/python.R
\name{cds_python_to_r}
\alias{cds_python_to_r}
\title{Translate Python Climate Data Store request to R request}
\usage{
cds_python_to_r(text, ...)
}
\arguments{
\item{text}{A \code{character} string containing the Python code copied
from a dataset download website \url{https://cds.climate.copernicus.eu/datasets}.
When missing, the function will use any text on the system clipboard.
This means that you can copy the API request code from the website to the
clipboard, then call this function without arguments.}

\item{...}{Ignored}
}
\value{
A named \code{list} that can be used as input for the functions
\code{cds_submit_job()} and \code{cds_estimate_costs()}
}
\description{
When looking for a dataset on \url{https://cds.climate.copernicus.eu/datasets},
you have the option to copy the API request code to the clipboard. However,
this is Python code and cannot be used directly in this package.
Use this function to translate the code to a request that can be handled by
the package. For more details see \code{vignette("translate")}
}
\examples{
python_code <-
"import cdsapi
dataset = \"reanalysis-era5-land\"
request = {
  \"variable\": [\"2m_dewpoint_temperature\"],
  \"year\": \"2025\",
  \"month\": \"01\",
  \"day\": [\"01\"],
  \"time\": [
    \"00:00\", \"01:00\", \"02:00\",
    \"03:00\", \"04:00\", \"05:00\",
    \"06:00\", \"07:00\", \"08:00\",
    \"09:00\", \"10:00\", \"11:00\",
    \"12:00\", \"13:00\", \"14:00\",
    \"15:00\", \"16:00\", \"17:00\",
    \"18:00\", \"19:00\", \"20:00\",
    \"21:00\", \"22:00\", \"23:00\"
  ],
  \"data_format\": \"netcdf\",
  \"download_format\": \"unarchived\"
}

client = cdsapi.Client()
client.retrieve(dataset, request).download()
"

cds_python_to_r(python_code)
}
\seealso{
Other helper-functions:
\code{\link[=cds_build_request]{cds_build_request()}},
\code{\link[=cds_cite_dataset]{cds_cite_dataset()}},
\code{\link[=cds_dataset_form]{cds_dataset_form()}},
\code{\link[=cds_estimate_costs]{cds_estimate_costs()}}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cite.R
\name{cds_cite_dataset}
\alias{cds_cite_dataset}
\title{Cite a dataset}
\usage{
cds_cite_dataset(dataset, ...)
}
\arguments{
\item{dataset}{The name of a dataset to be cited.}

\item{...}{Ignored}
}
\value{
Returns a \code{BibEntry}-class object, with citation details for the requested
dataset
}
\description{
Use this function to obtain citation details for a specific dataset
}
\examples{
if (interactive()) {
  cds_cite_dataset("reanalysis-era5-pressure-levels")
}
}
\seealso{
Other helper-functions:
\code{\link[=cds_build_request]{cds_build_request()}},
\code{\link[=cds_dataset_form]{cds_dataset_form()}},
\code{\link[=cds_estimate_costs]{cds_estimate_costs()}},
\code{\link[=cds_python_to_r]{cds_python_to_r()}}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token.R
\name{cds_get_token}
\alias{cds_get_token}
\alias{cds_set_token}
\title{Get or set a Climate Data Service API key}
\usage{
cds_get_token(...)

cds_set_token(token, method = c("option", "sysenv"), ...)
}
\arguments{
\item{...}{Ignored}

\item{token}{The API key you wish to set as an R option or to an environment variable.}

\item{method}{Method to store the API key. Should be either \code{"option"} (default) or
\code{"sysenv"}.}
}
\value{
\code{cds_get_token()} will return an API key token if it has been set. If
it is not set it will return an empty string: \code{""}. \code{cds_set_token()} will
return \code{NULL} invisibly.
}
\description{
Many of the Climate Data Services features require a personal
Application Programming Interface (API) key. This function
will get a token that was previously stored (\code{cds_set_token()}),
and can be used throughout the R session.
}
\details{
To use an API key, you first need to get one from the
Climate Data Service. You can do so, by creating an account
and then initialise the key at \url{https://cds.climate.copernicus.eu/profile}.

There are different locations where the key can be stored and where
\code{cds_get_token()} will look. It will look for the first successful value of (in order):
the environment variable named \code{"CDSAPI_KEY"}, the R \code{getOption()} named \code{"CDSAPI_KEY"},
the environment variable named \code{"ECMWF_DATASTORES_KEY"}, and the R \code{getOption()} named
\code{"ECMWF_DATASTORES_KEY"}.

You can set the key at the start of each R session with \code{cds_set_token()}. If you
want a persistent solution, you can add the environment variable (with names shown above)
to your system. Or you can add the option (with the names shown above) to your
\code{".profile"} file. This will help you obscure your sensitive account information in your R script.
}
\examples{
if (interactive() && !cds_token_works()) {
  cds_set_token("this-is-a-dummy-token", "option")
}

if (interactive()) {
  cds_get_token()
}
}
\seealso{
Other authentication-functions:
\code{\link[=cds_account_metrics]{cds_account_metrics()}},
\code{\link[=cds_check_authentication]{cds_check_authentication()}},
\code{\link[=cds_get_account]{cds_get_account()}}
}
\concept{authentication-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jobs.R
\name{cds_delete_job}
\alias{cds_delete_job}
\title{Delete/cancel jobs submitted to the Climate Data Service}
\usage{
cds_delete_job(job_id, ..., token = cds_get_token())
}
\arguments{
\item{job_id}{Hexadecimal code used as identifier of a job. Identifies the job
to be cancelled.}

\item{...}{Ignored}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
Returns a \code{data.frame} with information about the cancelled job.
}
\description{
When you regret submitting a job, you can cancel it by calling this function.
}
\examples{
if (interactive() && cds_token_works()) {
  job <- cds_submit_job(
      dataset        = "reanalysis-era5-pressure-levels",
      variable       = "geopotential",
      product_type   = "reanalysis",
      area           = c(n = 55, w = -1, s = 50, e = 10),
      year           = "2024",
      month          = "03",
      day            = "01",
      pressure_level = "1000",
      data_format    = "netcdf"
    )
  cds_delete_job(job$jobID, tempdir())
}
}
\seealso{
Other job-functions:
\code{\link[=cds_download_jobs]{cds_download_jobs()}},
\code{\link[=cds_job_results]{cds_job_results()}},
\code{\link[=cds_list_jobs]{cds_list_jobs()}},
\code{\link[=cds_submit_job]{cds_submit_job()}}
}
\concept{job-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token.R
\name{cds_get_account}
\alias{cds_get_account}
\title{Get account details}
\usage{
cds_get_account(token = cds_get_token(), ...)
}
\arguments{
\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}

\item{...}{Ignored}
}
\value{
Returns a named \code{list} with account details.
}
\description{
Retrieve account details associated with the provided \code{token}
}
\examples{
if (interactive() && cds_token_works()) {
  cds_get_account()
}
}
\seealso{
Other authentication-functions:
\code{\link[=cds_account_metrics]{cds_account_metrics()}},
\code{\link[=cds_check_authentication]{cds_check_authentication()}},
\code{\link[=cds_get_token]{cds_get_token()}}
}
\concept{authentication-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieve.R
\name{cds_build_request}
\alias{cds_build_request}
\title{Prepare a request for downloading a dataset}
\usage{
cds_build_request(dataset, ...)
}
\arguments{
\item{dataset}{The dataset name to be used for setting up a request.}

\item{...}{Parameters for subsetting the dataset. Use \code{cds_dataset_form()} to inquiry
which parameters and parameter values are available for a specific dataset.
If left blank it will take default parameter values.}
}
\value{
Returns a named list, which can be used to submit a job (\code{cds_submit_job()})
or inquiry its cost (\code{cds_estimate_costs()}).
}
\description{
This function is used by \code{cds_estimate_costs()} and \code{cds_submit_job()}
to subset a dataset before downloading. It will also help you to explore
which parameters are available for subsetting.
}
\examples{
if (interactive()) {
  cds_build_request(
    dataset        = "reanalysis-era5-pressure-levels",
    variable       = "geopotential",
    product_type   = "reanalysis",
    area           = c(n = 55, w = -1, s = 50, e = 10),
    year           = "2024",
    month          = "03",
    day            = "01",
    pressure_level = "1000",
    data_format    = "netcdf"
  )
}
}
\seealso{
Other helper-functions:
\code{\link[=cds_cite_dataset]{cds_cite_dataset()}},
\code{\link[=cds_dataset_form]{cds_dataset_form()}},
\code{\link[=cds_estimate_costs]{cds_estimate_costs()}},
\code{\link[=cds_python_to_r]{cds_python_to_r()}}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/retrieve.R
\name{cds_estimate_costs}
\alias{cds_estimate_costs}
\title{Check the cost of a request against your quota}
\usage{
cds_estimate_costs(dataset, ..., token = cds_get_token())
}
\arguments{
\item{dataset}{A dataset name to be inspected}

\item{...}{Parameters passed on to \code{cds_build_request()}}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
Returns a named list indicating the available quota and
the estimated cost for a request specified with \code{...}-arguments.
}
\description{
Each account has a limit to the amount of data that can be downloaded.
Use this function to check if a request exceeds your quota.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_estimate_costs(
    dataset        = "reanalysis-era5-pressure-levels",
    variable       = "geopotential",
    product_type   = "reanalysis",
    area           = c(n = 55, w = -1, s = 50, e = 10),
    year           = "2024",
    month          = "03",
    day            = "01",
    pressure_level = "1000",
    data_format    = "netcdf"
  )
  
  cds_estimate_costs(dataset = "reanalysis-era5-pressure-levels")
}
}
\seealso{
Other helper-functions:
\code{\link[=cds_build_request]{cds_build_request()}},
\code{\link[=cds_cite_dataset]{cds_cite_dataset()}},
\code{\link[=cds_dataset_form]{cds_dataset_form()}},
\code{\link[=cds_python_to_r]{cds_python_to_r()}}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stars.R
\name{cds_starred}
\alias{cds_starred}
\alias{cds_assign_star}
\alias{cds_remove_star}
\title{Add or remove a star to a dataset, or list starred datasets}
\usage{
cds_starred(..., token = cds_get_token())

cds_assign_star(dataset, ..., token = cds_get_token())

cds_remove_star(dataset, ..., token = cds_get_token())
}
\arguments{
\item{...}{Ignored}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}

\item{dataset}{Name of the dataset to assign a star to, or remove it from.}
}
\value{
In case of \code{cds_assign_star()} returns the name of the starred
dataset. In case of \code{cds_starred()} a vector of names of starred datasets.
In case of \code{cds_remove_star()} returns \code{NULL} invisibly.
}
\description{
Use stars to keep track of your favourite datasets. Use these
functions to add or remove a star.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_assign_star("reanalysis-carra-single-levels")
  cds_starred()
  cds_remove_star("reanalysis-carra-single-levels")
}
}
\seealso{
Other exploration-functions:
\code{\link[=cds_catalogue_vocabulary]{cds_catalogue_vocabulary()}},
\code{\link[=cds_list_datasets]{cds_list_datasets()}}
}
\concept{exploration-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CopernicusClimate-package.R
\docType{package}
\name{CopernicusClimate-package}
\alias{CopernicusClimate}
\alias{CopernicusClimate-package}
\title{CopernicusClimate: Search Download and Handle Data from Copernicus Climate Data Service}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Subset and download data from EU Copernicus Climate Data Service: \url{https://cds.climate.copernicus.eu/}. Import information about the Earth's past, present and future climate from Copernicus into R without the need of external software.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://pepijn-devries.github.io/CopernicusClimate/}
  \item \url{https://github.com/pepijn-devries/CopernicusClimate/}
  \item Report bugs at \url{https://github.com/pepijn-devries/CopernicusClimate/issues}
}

}
\author{
\strong{Maintainer}: Pepijn de Vries \email{pepijn.devries@outlook.com} (\href{https://orcid.org/0000-0002-7961-6646}{ORCID})

Authors:
\itemize{
  \item Pepijn de Vries \email{pepijn.devries@outlook.com} (\href{https://orcid.org/0000-0002-7961-6646}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/jobs.R
\name{cds_list_jobs}
\alias{cds_list_jobs}
\title{List jobs submitted to the Climate Data Service}
\usage{
cds_list_jobs(
  job_id = NULL,
  status = NULL,
  limit = 50,
  ...,
  token = cds_get_token()
)
}
\arguments{
\item{job_id}{The id of a specific job, if you want the results for that job. If \code{NULL}
(default) it is ignored.}

\item{status}{Only return jobs with the status stated by this argument. Default is \code{NULL}
meaning that jobs with any status are returned. Should be any of \code{"accepted"}, \code{"running"},
\code{"successful"}, \code{"failed"}, or \code{"rejected"}.}

\item{limit}{Use to limit the number of listed results. Defaults to \code{50}.}

\item{...}{Ignored}

\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}
}
\value{
Returns a \code{data.frame} of submitted jobs.
}
\description{
Once submitted with \code{cds_submit_job()} you can check the status of the job with
this function. You can list all available jobs, or specific jobs.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_list_jobs()
}
}
\seealso{
Other job-functions:
\code{\link[=cds_delete_job]{cds_delete_job()}},
\code{\link[=cds_download_jobs]{cds_download_jobs()}},
\code{\link[=cds_job_results]{cds_job_results()}},
\code{\link[=cds_submit_job]{cds_submit_job()}}
}
\concept{job-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/licenses.R
\name{cds_list_licences}
\alias{cds_list_licences}
\title{Obtain a list of licences that can be accepted}
\usage{
cds_list_licences(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} of available licences that can be accepted.
}
\description{
Some datasets require you to accept specific licences. This function will provide
an overview of all licences associated with datasets and can be accepted.
}
\examples{
if (interactive()) {
  cds_list_licences()
}
}
\seealso{
Other licences:
\code{\link[=cds_accept_licence]{cds_accept_licence()}},
\code{\link[=cds_accepted_licences]{cds_accepted_licences()}}
}
\concept{licences}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/token.R
\name{cds_check_authentication}
\alias{cds_check_authentication}
\alias{cds_token_works}
\title{Check if authentication works with a specific token}
\usage{
cds_check_authentication(token = cds_get_token(), ...)

cds_token_works(token = cds_get_token(), ...)
}
\arguments{
\item{token}{An API key to be used for authentication. Will use
\code{cds_get_token()} by default.}

\item{...}{Ignored}
}
\value{
\code{cds_check_authentication()} will return
some account information when successful but throws an error if
it is not. In contrast \code{cds_token_works()} returns a \code{logical}
value and will not throw an error upon failure.
}
\description{
Checks if the specified API key can be used for authentication at
the Climate Data Service.
}
\examples{
if (interactive() && cds_token_works()) {
  cds_check_authentication()
}
}
\seealso{
Other authentication-functions:
\code{\link[=cds_account_metrics]{cds_account_metrics()}},
\code{\link[=cds_get_account]{cds_get_account()}},
\code{\link[=cds_get_token]{cds_get_token()}}
}
\concept{authentication-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/catalogue.R
\name{cds_catalogue_vocabulary}
\alias{cds_catalogue_vocabulary}
\title{List catalogue vocabulary}
\usage{
cds_catalogue_vocabulary(...)
}
\arguments{
\item{...}{Ignored}
}
\value{
Returns a \code{data.frame} of keyword vocabulary used by the catalogue.
}
\description{
The catalogue uses a specific vocabulary for keywords. This function
produces an overview.
}
\examples{
if (interactive()) {
  cds_catalogue_vocabulary()
}
}
\seealso{
Other exploration-functions:
\code{\link[=cds_list_datasets]{cds_list_datasets()}},
\code{\link[=cds_starred]{cds_starred()}}
}
\concept{exploration-functions}
